from typing import Dict, Iterator, List

import requests
from requests.models import Response


class WskClient:
    DEFAULT_NS = 'guest'

    def __init__(self,
                 url: str,
                 username: str,
                 password: str,
                 ns: str=None) -> None:
        self.auth = (username, password)
        self.ns = ns if ns else self.DEFAULT_NS
        self.url = '{}/api/v1/namespaces/{}'.format(url, self.ns)

    @classmethod
    def from_config(cls, config: Dict) -> "WskClient":
        ow: Dict = config['openwhisk']
        return WskClient(
            url=ow['endpoint'],
            username=ow['auth']['username'],
            password=ow['auth']['password'],
            ns=ow['namespace'],
        )

    def create_action(self,
                      name: str,
                      runtime: str,
                      code: str,
                      main: str=None) -> Response:
        main = main if main else 'main'
        url = f'{self.url}/actions/{name}'
        r = requests.put(
            url,
            auth=self.auth,
            json={
                'namespace': self.ns,
                'name': name,
                'exec': {
                    'kind': runtime,
                    'code': code,
                    'main': main,
                },
                'publish': True,
            },
            params={'overwrite': 'true'},
        )
        return r

    def create_trigger(self, name: str) -> Response:
        url = f'{self.url}/triggers/{name}'
        r = requests.put(
            url,
            auth=self.auth,
            json={
                'namespace': self.ns,
                'name': name,
                'publish': True,
            },
            params={'overwrite': 'true'},
        )
        return r

    def create_rule(self,
                    name: str,
                    trigger_name: str,
                    action_name: str) -> Response:
        url = f'{self.url}/rules/{name}'
        r = requests.put(
            url,
            auth=self.auth,
            json={
                'name': name,
                'trigger': f'/_/{trigger_name}',
                'action': f'/_/{action_name}',
                'status': 'active',
                'publish': True,
            },
            params={'overwrite': 'true'},
        )
        return r
    
    def delete_action(self, name: str) -> Response:
        url = f'{self.url}/actions/{name}'
        r = requests.delete(url, auth=self.auth)
        return r
    
    def delete_trigger(self, name: str) -> Response:
        url = f'{self.url}/triggers/{name}'
        r = requests.delete(url, auth=self.auth)
        return r
    
    def delete_rule(self, name: str):
        url = f'{self.url}/rules/{name}'
        r = requests.delete(url, auth=self.auth)
        return r

    def fire_trigger(self, name: str, args: Dict=None) -> Response:
        args = args if args else dict()
        url = f'{self.url}/triggers/{name}'
        r = requests.post(
            url,
            auth=self.auth,
            json=args,
            params={
                'blocking': False,
                'result': False,
            },
        )
        return r
    
    def invoke_action(self, name: str, args: Dict=None) -> Response:
        args = args if args else dict()
        url = f'{self.url}/actions/{name}'
        r = requests.post(
            url,
            auth=self.auth,
            json=args,
            params={
                'blocking': False,
                'result': False,
            },
        )
        return r
    
    def list_actions(self, limit=200, skip=0) -> Response:
        """Returns list of action names (max limit: 200)."""
        url = f'{self.url}/actions'
        r = requests.get(
            url,
            auth=self.auth,
            params={
                'limit': limit,
                'skip': skip,
            },
        )
        return r
    
    def list_triggers(self, limit=200, skip=0) -> Response:
        """Returns list of trigger names (max limit: 200)."""
        url = f'{self.url}/triggers'
        r = requests.get(
            url,
            auth=self.auth,
            params={
                'limit': limit,
                'skip': skip,
            },
        )
        return r
    
    def list_rules(self, limit=200, skip=0) -> Response:
        """Returns list of rule names (max limit: 200)."""
        url = f'{self.url}/rules'
        r = requests.get(
            url,
            auth=self.auth,
            params={
                'limit': limit,
                'skip': skip,
            },
        )
        return r
    
    def action_paginator(self) -> "WskPaginator":
        return WskPaginator(client=self, resource='action')
    
    def trigger_paginator(self) -> "WskPaginator":
        return WskPaginator(client=self, resource='trigger')

    def rule_paginator(self) -> "WskPaginator":
        return WskPaginator(client=self, resource='rule')


class WskPaginator:
    def __init__(self, client: WskClient, resource: str) -> None:
        self.client = client

        if resource == 'action':
            self.list_method = self.client.list_actions
        elif resource == 'trigger':
            self.list_method = self.client.list_triggers
        elif resource == 'rule':
            self.list_method = self.client.list_rules
        else:
            raise AttributeError(
                f'No paginator found for resource {resource}.')
    
    def paginate(self) -> Iterator[str]:
        limit = 200
        skip = 0
        response = self.list_method(limit=limit, skip=skip)
        resources = [rsc['name'] for rsc in response.json()]
        while len(resources) == 200:
            for action in resources:
                yield action
            skip += 200
            resources = self.list_method(limit=limit, skip=skip)
        if len(resources) > 0:
            for action in resources:
                yield action


import pathlib
from typing import Dict, Union

from raise_me.models import Deployment

from .yaml_parser import YAMLParser
from .event import RaiseEventParser


class DeploymentParser:
    @classmethod
    def from_yaml(cls, path: Union[str, pathlib.Path]) -> Deployment:
        yaml_: Dict = YAMLParser.parse(path=path)
        
        if 'events' not in yaml_:
            raise SyntaxError('"Events" configuration not found.')

        if not isinstance(yaml_['events'], dict):
            raise SyntaxError('"events" configuration not found.')
        
        return Deployment(
            events=RaiseEventParser.parse_events(events=yaml_['events']))


import base64
import io
import pathlib
from typing import Dict, Union

from ruamel.yaml import YAML

from raise_me.parser import YAMLParser


class GCPWorkflowAdapter:
    def __init__(self,
                 config: Dict,
                 workflow_path: Union[str, pathlib.Path]) -> None:
        self.conf: Dict = config
        self.workflow: Dict = YAMLParser.parse(path=workflow_path)
    
    def to_pulumi_str(self, trigger_name: str) -> str:
        ow_url: str = '{}/api/v1/namespaces/{}'.format(
            self.conf['openwhisk']['endpoint'],
            self.conf['openwhisk']['namespace'],
        )
        for step in self.workflow['main']['steps']:
            if 'init_openwhisk_vars' in step:
                for var in step['init_openwhisk_vars']['assign']:
                    if 'TRIGGER_NAME' in var:
                        var['TRIGGER_NAME'] = trigger_name
                    elif 'OW_URL' in var:
                        var['OW_URL'] = ow_url
            
            if 'fire_openwhisk_trigger' in step:
                ow_auth: Dict[str, str] = self.conf['openwhisk']['auth']
                headers = step['fire_openwhisk_trigger']['args']['headers']
                headers['Authorization'] = 'Basic {}'.format(
                    base64.b64encode(
                        '{}:{}'.format(
                            ow_auth['username'],
                            ow_auth['password'],
                        ).encode()
                    ).decode('ascii'),
                )
        
        yaml = YAML()
        buffer = io.StringIO()
        yaml.dump(data=self.workflow, stream=buffer)
        workflow_str = buffer.getvalue()
        buffer.close()
        return workflow_str


from typing import Optional

import typer

from raise_me import RAISE_CONFIG_PATH, RAISE_DEPLOYMENT_PATH
from raise_me.build import OpenwhiskBuilder
from raise_me.parser import ConfigParser, DeploymentParser
from raise_me.wsk import WskClient


app = typer.Typer()


@app.command()
def up(config_path: Optional[str] = None, events_path: Optional[str] = None):
    conf_path = RAISE_CONFIG_PATH if config_path is None else config_path
    events_path = RAISE_DEPLOYMENT_PATH if events_path is None else events_path
    
    typer.echo('Creating RaiseMe OpenWhisk resources...')
    typer.echo(f' Configuration file: {conf_path}')
    typer.echo(f' Deployment file: {events_path}')
    
    OpenwhiskBuilder(
        wsk_client=WskClient.from_config(
            config=ConfigParser.from_yaml(path=conf_path)
    )).create_resources(
        deployment=DeploymentParser.from_yaml(path=events_path))

    typer.echo('OpenWhisk resources created!')


@app.command()
def destroy(config_path: Optional[str] = None):
    conf_path = RAISE_CONFIG_PATH if config_path is None else config_path

    typer.echo('Destroying RaiseMe OpenWhisk resources...')

    OpenwhiskBuilder(
        wsk_client=WskClient.from_config(
            config=ConfigParser.from_yaml(path=conf_path)
    )).destroy_resources()

    typer.echo('OpenWhisk resources destroyed!')


from typing import Dict

from .filter_ import FilterParser
from raise_me import PROVIDERS
from raise_me.models import EventSource


class EventSourceParser:
    VALID_KEYS = (
        'provider', 
        'filters',
    )

    @classmethod
    def from_dict(cls, source: Dict) -> EventSource:
        return EventSource(
            provider=source['provider'],
            filters=source['filters'],
        )

    @classmethod
    def valid(cls, source: Dict) -> bool:
        """Validates source syntax in terms of key values and types."""
        valid_keys = 'provider' in source and \
            all((key in cls.VALID_KEYS for key in source.keys()))

        if not valid_keys:
            return False
        
        if not isinstance(source['provider'], str) or \
                source['provider'] not in PROVIDERS:
            return False
        
        # Validate 'filters' is a non-empty List[str].
        if 'filters' in source.keys():
            if not FilterParser.valid(filters=source['filters'], 
                    provider=source['provider']):
                return False
        
        return True


__version__ = '0.0.1'

import pathlib


PROJECT_PATH = pathlib.Path(__file__).resolve().parent.parent
TESTS_PATH = PROJECT_PATH / 'tests'
RESOURCES_PATH = PROJECT_PATH / 'raise_me' / 'resources'

RAISE_DEPLOYMENT_PATH = PROJECT_PATH / 'raise-events.yaml'
RAISE_CONFIG_PATH = PROJECT_PATH / 'raise-config.yaml'

# Providers available.
PROVIDERS = (
    'aws',
    'gcp',
)



from typing import Dict, List

import pulumi
import pulumi_gcp as gcp

from . import WORKFLOW_TRIGGER_PATH
from raise_me.identity.openwhisk import OWResourceIdentifier
from raise_me.models import RaiseEvent
from raise_me.parser import FilterParser
from raise_me.util import GCPWorkflowAdapter


class GCPCloud:
    def __init__(self, config: Dict) -> None:
        self.conf = config
        self.project_id = config['gcp']['project-id']
        self.region = config['gcp']['region']
        self.wf_helper = GCPWorkflowAdapter(
            config=config,
            workflow_path=WORKFLOW_TRIGGER_PATH,
        )

    def update_stack(self, events: List[RaiseEvent]):
        if len(events) > 0:
            for event in events:
                service_account = gcp.serviceaccount.Account(
                    resource_name='raise-me_service-account',
                    account_id='raise-me-service-account',
                    display_name='Raise-Me Service Account',
                )

                workflows_iam_bind = gcp.projects.IAMMember(
                    resource_name='raise-me_iam-bind_workflows-invoker',
                    project=self.project_id,
                    member=service_account.email.apply(
                        lambda email: f'serviceAccount:{email}'),
                    role='roles/workflows.invoker',
                )
                eventrac_iam_bind = gcp.projects.IAMMember(
                    resource_name='raise-me_iam-bind_eventrac-eventReceiver',
                    project=self.project_id,
                    member=service_account.email.apply(
                        lambda email: f'serviceAccount:{email}'),
                    role='roles/eventarc.eventReceiver',
                )

                workflow = gcp.workflows.Workflow(
                    resource_name=f'raise-me_workflow_{event.logical_name}',
                    region=self.region,
                    description=f'Route raise-me event {event.logical_name}' \
                        'to fire OpenWhisk Trigger.',
                    service_account=service_account.id,
                    project=self.project_id,
                    source_contents=self.wf_helper.to_pulumi_str(
                        trigger_name=OWResourceIdentifier.trigger(
                            event_name=event.logical_name,
                        ),
                    ),
                )

                filters: List[Dict[str, str]] = FilterParser \
                    .to_eventrac_filters(filters=event.source.filters)

                trigger = gcp.eventarc.Trigger(
                    resource_name='raise-me-eventrac-trigger-{}'.format(
                        event.logical_name,
                    ),
                    opts=pulumi.ResourceOptions(depends_on=[
                        workflows_iam_bind, eventrac_iam_bind
                    ]),
                    location=self.region,
                    project=self.project_id,
                    matching_criterias=[
                        gcp.eventarc.TriggerMatchingCriteriaArgs(
                            attribute=f['attribute'],
                            value=f['value'],
                        )
                        for f in filters
                    ],
                    destination=gcp.eventarc.TriggerDestinationArgs(
                        workflow=workflow.name,
                    ),
                    service_account=service_account.name,
                )


import os
import requests
from typing import Dict, List

# Variable values added when mediator action deployed.
HTTP_TARGETS: List[Dict]
AUTH: tuple


def main(args: dict) -> dict:
    event: Dict = args.get('event')

    apihost = os.getenv('__OW_API_HOST')
    ns = os.getenv('__OW_NAMESPACE')
    invoker_name = 'raise-http-invoker'
    invoker_url = '{}:443/api/v1/namespaces/{}/actions/{}'.format(
        apihost, ns, invoker_name)
    
    for target in HTTP_TARGETS:
        # Invoke 'raise-http-invoker' action.
        _ = requests.post(
            invoker_url,
            auth=AUTH,
            json={
                'event': event,
                'http_target': target,
            },
            params={
                'blocking': False,
                'result': False,
            },
        )

    return {}


from raise_me import RESOURCES_PATH


LAMBDA_RESOURCES_PATH = RESOURCES_PATH / 'aws' / 'lambda'
CLOUDEVENTS_LAYER_PATH = LAMBDA_RESOURCES_PATH / 'layers' \
    / 'cloudevents-1.6.1.zip'
REQUESTS_LAYER_PATH = LAMBDA_RESOURCES_PATH / 'layers' \
    / 'requests-2.28.1.zip'
LAMBDA_TRIGGER_PATH = LAMBDA_RESOURCES_PATH / 'functions' \
    / 'openwhisk_trigger.zip'
WORKFLOW_TRIGGER_PATH = RESOURCES_PATH / 'gcp' / 'workflows' \
    / 'openwhisk_trigger.yaml'


from abc import ABC, abstractmethod
from dataclasses import dataclass

from .http_method import HttpMethod


class EventTarget(ABC):
    @abstractmethod
    def json(self):
        raise NotImplementedError()


@dataclass(frozen=True)
class HttpTarget(EventTarget):
    method: HttpMethod
    url: str
    # body: Optional[dict]

    def __str__(self) -> str:
        return f'HttpTarget: {self.method}|{self.url}'

    def json(self):
        return {
            'method': self.method.value,
            'url': self.url,
        }


@dataclass(frozen=True)
class ActionTarget(EventTarget):
    name: str

    def __str__(self) -> str:
        return f'CustomTarget: {self.name}'
    
    def json(self):
        return {
            'name': self.name,
        }


import json
from typing import Dict, List, Union


class FilterParser:
    @classmethod
    def to_aws_event_pattern(cls, 
                             filters: List[str],
                            ) -> Dict[str, Union[str, List, Dict]]:
        filters = filters.copy()
        for index, f in enumerate(filters):
            sep_index = f.index(':')
            key, val = f[:sep_index], f[sep_index + 1:]
            filters[index] = (key, json.loads(val))
        return json.dumps({key: val for key, val in filters})
    
    @classmethod
    def to_eventrac_filters(cls, filters: List[str]) -> List[Dict[str, str]]:
        filters = filters.copy()
        for index, f in enumerate(filters):
            sep_index = f.index('=')
            key, val = f[:sep_index], f[sep_index + 1:]
            filters[index] = {'attribute': key, 'value': val}
        return filters

    @classmethod
    def valid(cls, filters: List[str], provider: str) -> bool:
        """Validates filters syntax in terms of accepted provider formats.
        
        AWS event patterns are composed by key/value pairs separated by ':'.
        GCP Eventrac filters are composed by key/value pairs separated by '='.
        """

        if not isinstance(filters, List) or \
                len(filters) == 0 or \
                not all([isinstance(f, str) for f in filters]):
            return False
        
        split_char = ':' if provider == 'aws' else '='
        for f in filters:
            split = f.split(split_char)
            if len(split) < 2 or (len(split[0]) == 0 or len(split[1]) == 0):
                return False
        
        return True


import json
import os
import requests

from cloudevents.http import CloudEvent
from cloudevents.conversion import to_structured


# TODO: Use SecretsManager instead of AUTH variable.


AUTH = (os.environ.get('OW_USER'), os.environ.get('OW_PASS'))
OW_URL = os.environ.get('OW_URL')
TRIGGER_NAME = os.environ.get('TRIGGER_NAME')


def handler(event, context):
    attributes = {
        "type": event['detail-type'].replace(" ", "."),
        "source": event['source'],
        "subject": event['source'],
        "id": event['id'],
        "time": event['time'],
    }
    data = {
        "account": event['account'],
        "region": event['region'],
        "resources": event['resources'],
        "detail": event['detail'],
    }

    event = CloudEvent(attributes, data)
    _, body = to_structured(event)
    body = {'event': json.loads(body.decode())}
    print(event)

    url = f'{OW_URL}/triggers/{TRIGGER_NAME}'
    res = requests.post(
        url,
        auth=AUTH,
        json=body,
        params={
            'blocking': False,
            'result': False,
        },
    )
    
    return {
        'statusCode': res.status_code,
        'body': {
            'ow_response': res.json(),
        },
    }


import json
from typing import Dict, List

import pulumi
import pulumi_aws as aws

from . import CLOUDEVENTS_LAYER_PATH, LAMBDA_TRIGGER_PATH, REQUESTS_LAYER_PATH
from raise_me.models import RaiseEvent
from raise_me.identity.openwhisk import OWResourceIdentifier
from raise_me.parser import FilterParser


class AWSCloud:
    def __init__(self, config: Dict) -> None:
        self.conf = config

    def update_stack(self, events: List[RaiseEvent]):
        if len(events) > 0:
            cloudevents_layer = aws.lambda_.LayerVersion(
                resource_name='raise-me_lambda_cloudevents-layer',
                compatible_runtimes=['python3.7', 'python3.8', 'python3.9'],
                code=pulumi.FileArchive(path=str(CLOUDEVENTS_LAYER_PATH)),
                layer_name='raise-me_{}'.format(
                    CLOUDEVENTS_LAYER_PATH.stem.replace('.', '_')),
            )

            requests_layer = aws.lambda_.LayerVersion(
                resource_name='raise-me_lambda_requests-layer',
                compatible_runtimes=['python3.7', 'python3.8', 'python3.9'],
                code=pulumi.FileArchive(path=str(REQUESTS_LAYER_PATH)),
                layer_name='raise-me_{}'.format(
                    REQUESTS_LAYER_PATH.stem.replace('.', '_')),
            )

            basic_exec_role = aws.iam.get_policy(
                arn='arn:aws:iam::aws:policy/service-role/' \
                    'AWSLambdaBasicExecutionRole'
            )

            lambda_role = aws.iam.Role(
                resource_name='raise-me_iam-role_lambda',
                assume_role_policy=json.dumps({
                    'Version': '2012-10-17',
                    'Statement': [{
                        'Effect': 'Allow',
                        'Action': 'sts:AssumeRole',
                        'Principal': {
                            'Service': 'lambda.amazonaws.com'
                        },
                    }],
                }),
                inline_policies=[
                    aws.iam.RoleInlinePolicyArgs(
                        name='cloudwatch-access',
                        policy=basic_exec_role.policy,
                    ),
                ]
            )
            
            for event in events:
                ow_conf: Dict = self.conf['openwhisk']

                event_lambda = aws.lambda_.Function(
                    resource_name='raise-me_lambda-function_{}'.format(
                        event.logical_name),
                    code=str(LAMBDA_TRIGGER_PATH),
                    role=lambda_role.arn,
                    handler='openwhisk_trigger.handler',
                    runtime='python3.8',
                    environment=aws.lambda_.FunctionEnvironmentArgs(
                        variables={
                            'TRIGGER_NAME': OWResourceIdentifier.trigger(
                                event_name=event.logical_name),
                            'OW_URL': '{}/api/v1/namespaces/{}'.format(
                                ow_conf['endpoint'],
                                ow_conf['namespace'],
                            ),
                            "OW_USER": ow_conf['auth']['username'],
                            'OW_PASS': ow_conf['auth']['password'],
                        },
                    ),
                    layers=[cloudevents_layer.arn, requests_layer.arn],
                )

                event_rule = aws.cloudwatch.EventRule(
                    resource_name=f'raise-me_event-rule_{event.logical_name}',
                    description='Route raise-me event ' \
                        f'"{event.logical_name}" to lambda function. ' \
                        'Converts event into CloudEvents format and fires ' \
                        'OpenWhisk Trigger.',
                    # event_bus_name=event_bus.name,
                    event_pattern=FilterParser.to_aws_event_pattern(
                        filters=event.source.filters),
                    name=f'raise-me_{event.logical_name}',
                    is_enabled=True,
                )

                allow_eventbridge = aws.lambda_.Permission(
                    resource_name='raise-me_allow-eventbridge-invoke-lambda',
                    action='lambda:InvokeFunction',
                    function=event_lambda.name,
                    principal='events.amazonaws.com',
                    source_arn=event_rule.arn,
                )

                event_target = aws.cloudwatch.EventTarget(
                    resource_name=f'raise-me_event-target_{event.logical_name}',
                    rule=event_rule.name,
                    # event_bus_name=event_bus.name,
                    arn=event_lambda.arn,
                )


from typing import Dict
from raise_me.exceptions import UnrecognizedTargetType
from raise_me.models.event import HttpMethod

from raise_me.models.event.target import ActionTarget, EventTarget, HttpTarget


class EventTargetParser:
    VALID_KEYS = (
        'http', 
        'action',
    )

    @classmethod
    def from_dict(cls, target: dict) -> EventTarget:
        t_type = list(target.keys())[0]
        if t_type == 'http':
            return HttpTargetParser.from_dict(http_target=target[t_type])
        elif t_type == 'action':
            return ActionTargetParser.from_dict(action_target=target[t_type])
            
        raise UnrecognizedTargetType(f'{target}')

    @classmethod
    def valid(cls, target: dict) -> bool:
        """Validates targets' syntax in terms of key values and types."""
        if isinstance(target, dict) and \
                len(target.keys()) == 1 and \
                list(target.keys())[0] in cls.VALID_KEYS:
            t_type = list(target.keys())[0]

            if not isinstance(target[t_type], dict):
                return False

            valid_target = HttpTargetParser.valid(http_target=target[t_type]) \
                if t_type == 'http' \
                else ActionTargetParser.valid(action_target=target[t_type])
            
            if not valid_target:
                return False
        else:
            return False
        return True


class HttpTargetParser:
    VALID_KEYS = (
        'method', 
        'url',
    )

    @classmethod
    def from_dict(cls, http_target: dict) -> HttpTarget:
        return HttpTarget(
            method=HttpMethod.from_str(val=http_target['method']),
            url=http_target['url'],
        )

    @classmethod
    def valid(cls, http_target: dict) -> bool:
        valid_keys = 'method' in http_target and \
            'url' in http_target and \
            all((key in cls.VALID_KEYS for key in http_target.keys()))
        
        if not valid_keys:
            return False
        if not isinstance(http_target['method'], str):
            return False
        if not isinstance(http_target['url'], str):
            return False
        return True


class ActionTargetParser:
    VALID_KEYS = (
        'name',
    )

    @classmethod
    def from_dict(cls, action_target: dict) -> ActionTarget:
        return ActionTarget(
            name=action_target['name'],
        )

    @classmethod
    def valid(cls, action_target: dict) -> bool:
        if 'name' not in action_target:
            return False
        if not isinstance(action_target['name'], str):
            return False
        return True


import pathlib
from typing import Dict, Union

from .yaml_parser import YAMLParser


class ConfigParser:
    @classmethod
    def from_yaml(cls, path: Union[str, pathlib.Path]) -> Dict:
        config: Dict = YAMLParser.parse(path=path)
        if not cls.valid(config=config):
            raise SyntaxError('Configuration file "raise-config.yaml" '
                'malformed.')
        return config
    
    @classmethod
    def valid(cls, config: Dict) -> bool:
        if not cls.valid_openwhisk(config=config):
            return False
        if not cls.valid_gcp(config=config):
            return False
        return True
    
    @classmethod
    def valid_openwhisk(cls, config: Dict) -> bool:
        if 'openwhisk' not in config or \
                not isinstance(config['openwhisk'], dict):
            return False
        
        ow = config['openwhisk']
        if 'namespace' not in ow or 'endpoint' not in ow or 'auth' not in ow:
            return False
        
        if not isinstance(ow['namespace'], str) or \
                not isinstance(ow['endpoint'], str) or \
                not isinstance(ow['auth'], dict):
            return False
        
        auth = ow['auth']
        if 'username' not in auth or 'password' not in auth:
            return False
        
        if not isinstance(auth['username'], str) or \
                not isinstance(auth['password'], str):
            return False
        
        if len(ow['namespace'].strip()) == 0 or \
                len(ow['endpoint'].strip()) == 0:
            return False
        
        if len(auth['username'].strip()) == 0 or \
                len(auth['password'].strip()) == 0:
            return False
        return True
    
    @classmethod
    def valid_gcp(cls, config: Dict) -> bool:
        if 'gcp' not in config or \
                not isinstance(config['gcp'], dict):
            return False
        
        gcp = config['gcp']
        if 'project-id' not in gcp or 'region' not in gcp:
            return False

        if not isinstance(gcp['project-id'], str) or \
                not isinstance(gcp['region'], str):
            return False
        
        if len(gcp['project-id'].strip()) == 0 or \
                len(gcp['region'].strip()) == 0:
            return False
        return True


from dataclasses import dataclass, field
from typing import List

from .event import RaiseEvent


@dataclass
class Deployment:
    events: List[RaiseEvent] = field(default_factory=list)

    def __post__init__(self):
        self.events.sort(key=lambda e: e.logical_name)

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, Deployment):
            return str(self) == str(__o)
        return False

    def __str__(self) -> str:
        return '{{Events: {}}}'.format([str(e) for e in self.events])


from dataclasses import dataclass
from typing import List

from .target import EventTarget
from .source import EventSource


@dataclass(frozen=True)
class RaiseEvent:
    logical_name: str
    source: EventSource
    targets: List[EventTarget]

    def __post_init__(self):
        self.targets.sort(key=lambda t: str(t))

    def __eq__(self, __o: object) -> bool:
        return (self.logical_name == __o.logical_name
            and self.source == __o.source 
            and set([str(t) for t in self.targets]) == 
                set([str(t) for t in __o.targets]))
    
    def __str__(self) -> str:
        return '{{{}: {{Source: \'{}\', Targets: {}}}}}'.format(
            self.logical_name,
            self.source,
            [str(t) for t in self.targets])


"""Module of exceptions that raise-me may raise."""


class InvalidEvent(SyntaxError):
    """Exception when an event from 'raise-events.yaml' is malformed."""
    pass


class UnrecognizedTargetType(Exception):
    """Exception when EventTarget type is not valid."""
    pass


class UnrecognizedHttpMethod(Exception):
    """Exception when http method is not valid."""
    pass


from typing import List

from raise_me.models import RaiseEvent
from raise_me.exceptions import InvalidEvent

from .source import EventSourceParser
from .target import EventTargetParser


class RaiseEventParser:
    @classmethod
    def from_dict(cls, name: str, meta: dict) -> RaiseEvent:
        return RaiseEvent(
            logical_name=name,
            source=EventSourceParser.from_dict(source=meta['source']),
            targets=[EventTargetParser.from_dict(target=t) 
                     for t in meta['targets']],
        )

    @classmethod
    def parse_events(cls, events: dict) -> List[RaiseEvent]:
        raise_events: List[RaiseEvent] = []

        for name, event_meta in events.items():
            # Validate syntax.
            if not isinstance(event_meta['source'], dict) or \
                    not EventSourceParser.valid(source=event_meta['source']):
                raise InvalidEvent('Source from event "{}" malformed: {}.'
                    .format(name, event_meta['source']))
            
            if not isinstance(event_meta['targets'], List) or \
                    len(event_meta['targets']) == 0:
                raise InvalidEvent('Targets from event "{}" malformed: {}.'
                    .format(name, event_meta['targets']))

            for target in event_meta['targets']:
                if not EventTargetParser.valid(target=target):
                    raise InvalidEvent('Target from event "{}" malformed: {}.'
                        .format(name, target))

            # Instantiate model.
            raise_events.append(cls.from_dict(name=name, meta=event_meta))
        
        return raise_events


from .config import ConfigParser
from .deployment import DeploymentParser
from .event import RaiseEventParser, EventSourceParser, EventTargetParser, \
    FilterParser
from .yaml_parser import YAMLParser


import requests
from typing import Dict


def main(args: dict) -> dict:
    event: Dict = args.get('event')
    http_target: Dict = args.get('http_target')

    http_method: str = http_target.get('method')
    url: str = http_target.get('url')
    
    if http_method == 'get':
        r = requests.get(url)
    else: # post
        r = requests.post(url, json=event)

    try:
        response = r.json()
    except ValueError as err:
        response = {'body': r.text}
    finally:
        print(response) # Log http response.
        return response


from . import HTTP_INVOKER_ACTION_NAME


class OWResourceIdentifier:
    @classmethod
    def http_invoker(cls) -> str:
        return HTTP_INVOKER_ACTION_NAME
    
    @classmethod
    def http_mediator(cls, event_name: str) -> str:
        return f'raise_mediator-{event_name}'
    
    @classmethod
    def trigger(cls, event_name: str) -> str:
        return f'raise_trigger-{event_name}'
    
    @classmethod
    def rule(cls, trigger_name: str, target_name: str) -> str:
        return f'raise_rule-{trigger_name}-{target_name}'


from dataclasses import dataclass
from typing import List


@dataclass(frozen=True)
class EventSource:
    provider: str
    filters: List[str]

    def __post_init__(self):
        self.filters.sort()

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, EventSource):
            return self.provider == __o.provider and \
                set(self.filters) == set(__o.filters)
        return False
    
    def __str__(self) -> str:
        return f'EventSource: {self.provider}|{self.filters}'


import pathlib
from typing import List, Union

from requests.models import Response

from raise_me.models import Deployment, HttpTarget, ActionTarget
from raise_me.identity.openwhisk import OWResourceIdentifier
from raise_me.wsk import WskClient
from raise_me.wsk import HTTP_MEDIATOR_PATH, HTTP_INVOKER_PATH


class OpenwhiskBuilder:
    def __init__(self, wsk_client: WskClient) -> None:
        self.wsk_client = wsk_client

    def create_resources(self, deployment: Deployment) -> None:
        """Create Openwhisk Actions, Triggers and Rules that represent the core
        of the event broker system. 

        External provider events will fire specific Triggers that will invoke 
        the corresponding Actions associated to that event to reach all targets
        declared in 'raise-events.yaml'.

        Logic description:
        - Deploy 'raise-http-invoker' action, receives (event, http_target).
        - Create event trigger (can be fired externally).
        - For each action_target:
            - Create rule to link trigger to action_target.
        - If at least 1 http_target, create http_mediator, which will invoke 
            the 'raise-http-invoker' action for every http_target.
            - For each http_target:
                - Create rule to link trigger to http_mediator.
        """        
        # Create 'raise-http-invoker' action by default.
        _ = self.create_http_invoker(
            name=OWResourceIdentifier.http_invoker(),
            wsk_client=self.wsk_client, 
            path=HTTP_INVOKER_PATH,
        )
        
        for event in deployment.events:
            targets_to_mediate: List[HttpTarget] = []
            targets_to_trigger: List[ActionTarget] = []

            for target in event.targets:
                if isinstance(target, HttpTarget):
                    targets_to_mediate.append(target)
                elif isinstance(target, ActionTarget):
                    targets_to_trigger.append(target)
            
            # Create a event trigger.
            trigger_name = OWResourceIdentifier.trigger(
                event_name=event.logical_name)
            _ = self.wsk_client.create_trigger(name=trigger_name)

            # Create a rule for every user-defined action_target.
            for target in targets_to_trigger:
                _ = self.wsk_client.create_rule(
                    name=OWResourceIdentifier.rule(
                        trigger_name=trigger_name,
                        target_name=target.name,
                    ),
                    trigger_name=trigger_name,
                    action_name=target.name,
                )
            
            if len(targets_to_mediate) > 0:
                # Create http-mediator event action and link to trigger.
                event_mediator_name = OWResourceIdentifier.http_mediator(
                    event_name=event.logical_name)
                _ = self.create_http_mediator(
                    name=event_mediator_name,
                    http_targets=targets_to_mediate,
                    wsk_client=self.wsk_client,
                    path=HTTP_MEDIATOR_PATH,
                )
                _ = self.wsk_client.create_rule(
                    name=OWResourceIdentifier.rule(
                        trigger_name=trigger_name,
                        target_name=event_mediator_name,
                    ),
                    trigger_name=trigger_name,
                    action_name=event_mediator_name,
                )

    def destroy_resources(self):
        """Destroy all actions, triggers and rules created by this class."""
        self.wsk_client.delete_action(name=OWResourceIdentifier.http_invoker())

        for action_name in self.wsk_client.action_paginator().paginate():
            if action_name.startswith('raise_mediator-'):
                self.wsk_client.delete_action(name=action_name)
        
        for rule_name in self.wsk_client.rule_paginator().paginate():
            if rule_name.startswith('raise_rule-'):
                self.wsk_client.delete_rule(name=rule_name)
        
        for trigger_name in self.wsk_client.trigger_paginator().paginate():
            if trigger_name.startswith('raise_trigger-'):
                self.wsk_client.delete_trigger(name=trigger_name)

    @classmethod
    def create_http_invoker(cls,
                            name: str,
                            wsk_client: WskClient,
                            path: Union[str, pathlib.Path]) -> Response:
        return wsk_client.create_action(
            name=name,
            runtime='python:3',
            code=open(path, 'r').read(),
            main='main',
        )
    
    @classmethod
    def create_http_mediator(cls,
                             name: str,
                             http_targets: List[HttpTarget],
                             wsk_client: WskClient,
                             path: Union[str, pathlib.Path]) -> Response:
        adapted_code = 'HTTP_TARGETS={}\nAUTH={}\n{}'.format(
            [target.json() for target in http_targets],
            wsk_client.auth,
            open(path, 'r').read(),
        )
        return wsk_client.create_action(
            name=name,
            runtime='python:3',
            code=adapted_code,
            main='main',
        )


import pathlib
from typing import Dict, Union

from .aws import AWSCloud
from .gcp import GCPCloud
from raise_me.models import Deployment
from raise_me.parser import ConfigParser, DeploymentParser


class CloudOrchestrator:
    def __init__(self, config_path: Union[str, pathlib.Path]) -> None:
        self.conf: Dict = ConfigParser.from_yaml(path=config_path)
        self.aws_cloud = AWSCloud(config=self.conf)
        self.gcp_cloud = GCPCloud(config=self.conf)
    
    def update_stack(self, events_path: Union[str, pathlib.Path]) -> None:
        deployment: Deployment = DeploymentParser.from_yaml(path=events_path)
        
        self.aws_cloud.update_stack(
            events=[event for event in deployment.events 
            if event.source.provider == 'aws'])
        
        self.gcp_cloud.update_stack(
            events=[event for event in deployment.events 
            if event.source.provider == 'gcp'])


from enum import Enum

from raise_me.exceptions import UnrecognizedHttpMethod


class HttpMethod(Enum):
    GET='get'
    POST='post'

    @classmethod
    def from_str(cls, val: str) -> "HttpMethod":
        if val == 'get': 
            return HttpMethod.GET
        elif val == 'post': 
            return HttpMethod.POST
        raise UnrecognizedHttpMethod(val)


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['raise_me',
 'raise_me.build',
 'raise_me.build.cloud',
 'raise_me.exceptions',
 'raise_me.identity',
 'raise_me.models',
 'raise_me.models.event',
 'raise_me.parser',
 'raise_me.parser.event',
 'raise_me.resources.aws.lambda.functions',
 'raise_me.util',
 'raise_me.util.gcp',
 'raise_me.wsk']

package_data = \
{'': ['*'],
 'raise_me': ['resources/aws/lambda/layers/*', 'resources/gcp/workflows/*']}

install_requires = \
['boto3>=1.24.50,<2.0.0',
 'cloudevents>=1.6.1,<2.0.0',
 'constructs>=10.1.85,<11.0.0',
 'pulumi-aws>=5.11.0,<6.0.0',
 'pulumi-gcp>=6.37.0,<7.0.0',
 'pulumi>=3.38.0,<4.0.0',
 'requests>=2.28.1,<3.0.0',
 'ruamel.yaml>=0.17.21,<0.18.0',
 'typer[all]>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['raise = raise_me.app:app']}

setup_kwargs = {
    'name': 'raise-me',
    'version': '0.0.1',
    'description': 'Serverless event broker which allows connecting native cloud events (AWS & GCP) to targets declaratively.',
    'long_description': '\n# raise-me\nServerless event broker which allows connecting native cloud events to targets declaratively.\n- The native cloud **events** can be raised from either *AWS* or *GCP*.\n- The **targets** can be either *HTTP endpoints* or [*OpenWhisk Actions*](https://github.com/apache/openwhisk/blob/master/docs/actions.md).\n\n\n### Example\nDeclare your event defintions in `raise-events.yaml`:\n```yaml\nevents:\n\n  s3-to-cloudfunction: # Event\'s logical name.\n    source:\n      provider: aws\n      filters: # AWS EventBridge event patterns.\n        - \'source: ["aws.s3"]\'\n        - \'detail-type: ["Object Created"]\'\n        - \'detail: {"bucket": {"name": ["my-bucket-name"]}}\'\n    targets:\n      - http:\n          method: post # Sends event data in CloudEvents format.\n          url: my-cloudfunction-url\n      - action:\n          name: my-openwhisk-action-name\n\n  cloudstorage-to-lambda:\n    source:\n      provider: gcp\n      filters: # GCP Eventrac filters.\n        - \'type=google.cloud.audit.log.v1.written\'\n        - \'serviceName=storage.googleapis.com\'\n        - \'methodName=storage.objects.create\'\n    targets:\n      - http:\n          method: get\n          url: my-lambda-url\n```\n\n---\n\n## Introduction\nThis solution enables declarative native cloud event routing (e.g. AWS S3\'s "Object Created") to either HTTP endpoints or user-defined serverless functions deployed as [Apache OpenWhisk](https://openwhisk.apache.org/) Actions.\n\nAn example use case could be triggering a [GCP Cloud Function](https://cloud.google.com/functions) when a file is uploaded to an [AWS S3](https://aws.amazon.com/s3/) bucket.\nSuch process is accomplished by creating necessary cloud resources that will route the event to the OpenWhisk deployment, which will then receive the event\'s data in the [CloudEvents](https://cloudevents.io/) format and forward it to the Cloud Function through an API call.\n\n---\n\n## Table of Contents\n- [Motivation](#motivation)\n- [Requirements](#requirements)\n- [Configuration](#configuration)\n  - [Pulumi](#pulumi)\n  - [Enable Google Services](#enable-google-services)\n- [Installation](#installation)\n- [Testing](#testing)\n- [Usage](#usage)\n  - [1. Prepare Inputs](#1-prepare-inputs)\n  - [2. Create Resources](#2-create-resources)\n  - [3. Cleanup](#3-cleanup)\n- [Architecture](#architecture)\n- [Contriburing](#architecture)\n- [License](#architecture)\n- [Future Features](#future-features)\n\n---\n\n## Motivation\nThe implementation alternatives for the previous example (trigger Cloud Function from S3 event) are countless. A straightforward approach could be linking a Lambda to the event, which is fairly simple within the AWS ecosystem, and communicate with the Cloud Function from there.\n\nAlthough this alternative may seem initially as a good approach, there are several design limitations. If we were to trigger 2 Cloud functions instead of 1, we would have to deploy a new Lambda with code containing logic to reach both Cloud functions.\nWhat if we want to reach 100 endpoints from that single event? Lambdas have a 15-minute time limit on their execution, so extending this event-based communication solution would have limitations in terms of:\n- ***Extensibility** – the number of endpoints would be limited by Lambda. Perhaps, we would have to create more Lambdas to handle different endpoints.*\n- ***Maintainability** – the logic at the Lambda layer might require frequent changes.*\n- ***Complexity** – the logic at the Lambda layer will increase.*\n- ***Portability** – this schema would only work for native AWS events, but not for other clouds, such as GCP.*\n\nThe solution that **raise-me** provides addresses all of these limitations by using serverless services and an abstraction layer that simplifies the management of the required infrastrcture that makes the event routing possible.\n\n---\n\n## Requirements\n- [Apache OpenWhisk](https://openwhisk.apache.org/) deployment - *where the event processors run; OpenWhisk has several [deployment options](https://openwhisk.apache.org/documentation.html#openwhisk_deployment), including Kubernetes.*\n- [Pulumi](https://www.pulumi.com/) - *used to create cloud resources.*\n- AWS Account & [AWS CLI](https://aws.amazon.com/cli/) - *optional, if AWS events are to be listened.*\n- GCP Account - *optional, if GCP events are to be listened.*\n\n---\n\n## Configuration\n#### Pulumi\n[Configure Pulumi to access your AWS account](https://www.pulumi.com/docs/get-started/aws/begin/) and/or [Google Cloud account](https://www.pulumi.com/docs/get-started/gcp/begin/). If you are just interested in one of those event sources, you can ignore the other.\n\nCreate a Pulumi project. You can use either a Python template, AWS or GCP. For example:\n```bash\n$ mikdir your-project-name\n$ pulumi new aws-python\n```\n\nActivate your pulumi project\'s virtual environment and install `raise-me` (available on PyPi):\n```bash\n$ source ./venv/bin/activate\n$ ./venv/bin/python.exe -m pip install raise-me\n```\n\nIn `Pulumi.<stack-name>.yaml`, make sure you have the appropriate configuration for the corresponding clouds:\n```yaml\nconfig:\n  aws:region: your-target-region\n  gcp:project: your-project-id\n```\n\n#### Enable Google Services\nEnable the following Google services:\n- [Eventrac](https://cloud.google.com/eventarc/docs/overview) - _eventrac triggers will be created._\n- [PubSub](https://cloud.google.com/pubsub) - _used internally by Eventrac._\n- [Workflows](https://cloud.google.com/workflows) - _workflows will be created._\n- [Cloud Logging](https://cloud.google.com/logging) - _provides more event options._\n\nIf you are interested in handling events from services that log into Cloud Logging, [enable the logging of those services](https://cloud.google.com/logging/docs/audit/configure-data-access#config-console). \n\n---\n\n## Installation\nRun (available on PyPi):\n```bash\n$ pip install raise-me\n```\nVerify installation:\n```bash\n$ raise --help\n```\n\n---\n\n## Testing\nInstall [Poetry](https://python-poetry.org/) if you don\'t have it already:\n```bash\n$ pip install poetry\n```\n\nClone the repository, create a virtual environment at the project level and activate it:\n```bash\n$ python3 -m venv my-venv\n$ source my-venv/bin/activate \n```\n\nInstall project dependencies using Poetry:\n```bash\n$ poetry install\n```\n\nModify the `tests/raise-config.yaml` and provide the connection details of the target OpenWhisk deployment (see [Usage](#usage) for more details). You can also modify the contents of `tests/raise-events.yaml`, but it is not necessary for these tests.\n\nAfter this, you can run the following integration tests that verify that the connection with the OpenWhisk deployment is successful:\n```bash\n$ pytest -v -m client # Tests OpenWhisk API client.\n$ pytest -v -m paginator # Tests paginator interface using the API client.\n$ pytest -v -m builder # Tests creation/destruction of raise-me OpenWhisk resources.\n```\n\n---\n\n## Usage\n\n#### 1. Prepare Inputs\nThere are 2 required inputs, namely `raise-config.yaml` and `raise-events.yaml`.\n\nThe `raise-config.yaml` file contains configuration settings that allow connecting to the OpenWhisk deployment and creating cloud resources. *(The auth field in the example contains default values for a basic unchanged OpenWhisk deployment.)*\n\n```yaml\nopenwhisk:\n  namespace: guest # Storage location of Openwhisk resources.\n  endpoint: https://my-endpoint:443 # Openwhisk deployment endpoint.\n  auth: # Authentication for REST API.\n    username: 23bc46b1-71f6-4ed5-8c54-816aa4f8c502\n    password: 123zO3xZCLrMN6v2BKK1dXYFpXlPkccOFqm12CdAsMgRU4VrNZ9lyGVCGuMDGIwP\n\ngcp: # Used for cloud resource creation.\n  project-id: my-project-id\n  region: my-region # Should match pulumi config.\n```\n\nThe `raise-events.yaml` will contain the event definitions. Each event definition is composed by a Source and a list of Targets.\nTo specify the desired events, we use each cloud\'s selection strategy; event patterns for AWS and Eventrac filters for GCP.\n\n```yaml\nevents:\n\n  s3-to-cloudfunction: # Your event\'s logical name.\n    source:\n      provider: aws\n      filters: # AWS EventBridge event patterns.\n        - \'source: ["aws.s3"]\'\n        - \'detail-type: ["Object Created"]\'\n        - \'detail: {"bucket": {"name": ["my-bucket-name"]}}\'\n    targets:\n      - http:\n          method: post # Sends event data in CloudEvents format.\n          url: my-cloudfunction-url\n      - action:\n          name: my-openwhisk-action-name\n\n  cloudstorage-to-lambda:\n    source:\n      provider: gcp\n      filters: # GCP Eventrac filters.\n        - \'type=google.cloud.audit.log.v1.written\'\n        - \'serviceName=storage.googleapis.com\'\n        - \'methodName=storage.objects.create\'\n    targets:\n      - http:\n          method: get\n          url: my-lambda-url\n```\n\n_Notes on the `raise-events.yaml`:_\n* _Each event definition must have only 1 `provider` and at least 1 target._\n* _The providers available are `aws` and `gcp`._\n* _Targets available include `http` and `action`._\n* _`http` targets must contain `method` and a `url`. Current available methods are `get` and `post`, where the later sends the event\'s data to the target._\n* _`action` targets must contain the `name` of the OpenWhisk action to trigger, which should be available within the namespace specified in the `raise-config.yaml`._\n* _OpenWhisk Actions are language-agnostic, so if you need more advanced event-handling logic, you can deploy your own actions and link them to the one or more events seamlessly!_\n\n\n#### 2. Create Resources\nOnce having the inputs ready, we just need to create:\n- Cloud resources that will route the events to our OpenWhisk deployment.\n- OpenWhisk resources that will handle the incoming events.\n\nCreate the OpenWhisk resources using the `raise` command:\n```bash\n$ raise up --config-path path/to/raise-config.yaml --events-path path/to/raise-events.yaml\n```\n\nCreate the cloud resources with Pulumi as you usually would:\n```bash\n~/pulumi/project/$ pulumi up \n```\n\nAnd done! You should have now the required infrastructure in place to route the specified events to their corresponding targets.\n\n\n#### 3. Cleanup\nTo delete the OpenWhisk resources created by **raise-me**, run:\n```bash\n$ raise destroy --config-path path/to/raise-config.yaml\n```\n\nDelete the cloud resources running:\n```bash\n~/pulumi/project/$ pulumi destroy\n```\n\n---\n\n### Architecture\nLet\'s consider the example given in the introduction and an OpenWhisk deployment in [AWS EKS](https://aws.amazon.com/eks/).\n\nTo forward S3 events to OpenWhisk, **raise-me** uses [AWS EventBridge](https://aws.amazon.com/eventbridge/). It creates an [EventRule](https://docs.aws.amazon.com/eventbridge/latest/userguide/eb-rules.html) that routes the event to a [Lambda](https://aws.amazon.com/lambda/) function that parses the event and forwards it to the OpenWhisk deployment by firing an [OpenWhisk Trigger](https://github.com/apache/openwhisk/blob/master/docs/triggers_rules.md#creating-triggers-and-rules). Such Trigger will then invoke the necessary Actions to reach the Cloud Function\'s endpoint.\n\n![Architecture](images/architecture-example-aws.PNG)\n\n_Note that:_\n- _The services used to route the events to our target, namely *EventBridge*, *Lambda* and *OpenWhisk*, are serverless and can scale seamlessly._\n- _The dotted arrows represent AWS managed communication, whereas the solid lines represent REST API calls implemented in the solution._\n- _Although OpenWhisk implements internal load balancing, a deployment in EKS will also include an Elastic Load Balancer ([docs](https://github.com/apache/openwhisk-deploy-kube/blob/master/docs/k8s-aws.md)). This component is omitted in the diagram as it is AWS-specific._\n\nConsidering the same OpenWhisk deployment, let\'s say that we would like to trigger a Lambda function when a file is uploaded to a Cloud Storage bucket. The architecture would look as follows:\n\n![Architecture](images/architecture-example-google.PNG)\n\n---\n\n## Contributing\nPull requests are welcome. For major changes, please open an issue first to discuss the change.\n\nPlease ensure tests are up to date.\n\n---\n\n## License\n[MIT](License)\n\n-----\n\n## Future Features\n- [ ] Thorough exception handling\n- [ ] Secure OpenWhisk API credentials\n  - [ ] AWS: SecretsManager\n  - [ ] GCP: SecretManager\n- [ ] Http request body/parameters\n- [ ] Internal API crendential storage (for HTTP targets)\n- [ ] Logging\n\n---\n\n### Roadmap\n- [X] Parser module\n- [X] Builder module\n    - [X] OpenWhisk resources\n    - [X] Cloud resources\n        - [X] AWS\n            - [X] Eventbridge\n            - [X] Lambda\n        - [X] GCP\n            - [X] Eventrac\n            - [X] Workflows\n- [X] CLI Tool\n- [X] Pytests (OpenWhisk)\n- [X] Add documentation.',
    'author': 'Fernando Ancona',
    'author_email': 'f.anconac@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ferancona/raise-me',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import pathlib
from typing import Dict, Union

from ruamel.yaml import YAML
from ruamel.yaml.constructor import DuplicateKeyError


class YAMLParser:
    @classmethod
    def parse(cls,  path: Union[str, pathlib.Path]) -> Dict:
        path = pathlib.Path(path).resolve()
        parser = YAML(typ='safe')
        try:
            with open(path, 'r') as f:
                yaml_: Dict = parser.load(f.read())
        except DuplicateKeyError as err:
            # Log.
            raise err
        
        return yaml_


from raise_me import PROJECT_PATH
from .client import WskClient, WskPaginator

WSK_MODULE_PATH = PROJECT_PATH / 'raise_me' / 'wsk'
HTTP_INVOKER_PATH = WSK_MODULE_PATH / 'http_invoker.py'
HTTP_MEDIATOR_PATH = WSK_MODULE_PATH / 'http_mediator.py'